﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Diagnostics;
using System.Configuration;
using ConsoleApplication1.SqlStringFactory.Configuration;

namespace ConsoleApplication1.SqlStringFactory
{
    /// <summary>
    /// Manages the primary database connection strings and SQL strings in the web.config.
    /// Can be called without specifying a key, allowing reflection to select the proper SQL string, based on the calling class and function name.
    /// </summary>
    public static class SqlFactory
    {
        private static SqlStringFactorySectionHandler sqlStringFactoryConfiguration = (SqlStringFactorySectionHandler)ConfigurationManager.GetSection("SqlStringFactory");

        /// <summary>
        /// Default connection string used to connect to the primary database.
        /// </summary>
        public static string ConnectionString
        {
            get
            {
                return ConfigurationManager.ConnectionStrings[sqlStringFactoryConfiguration.DefaultConnectionStringName].ConnectionString;
            }
        }

        #region Reflected Functions

        /// <summary>
        /// Returns the SQL string with the key matching the calling class and function name.
        /// </summary>
        /// <returns>SQL string</returns>
        public static string GetSql()
        {
            // Get the calling method via the stack trace.
            StackTrace stackTrace = new StackTrace();
            MethodBase reflectedMethod = stackTrace.GetFrame(1).GetMethod();

            // Get the class and method name via the reflected method.
            string className = reflectedMethod.DeclaringType.Name;
            string method = reflectedMethod.Name;

            return GetSql(className + "." + method);
        }

        /// <summary>
        /// Returns the SQL string with the key matching the calling class and function name. Passes in parameters.
        /// </summary>
        /// <param name="parameterList">List of parameter strings</param>
        /// <returns>SQL string</returns>
        public static string GetSql(List<string> parameterList)
        {
            // Get the calling method via the stack trace.
            StackTrace stackTrace = new StackTrace();
            MethodBase reflectedMethod = stackTrace.GetFrame(1).GetMethod();

            // Get the class and method name via the reflected method.
            string className = reflectedMethod.DeclaringType.Name;
            string method = reflectedMethod.Name;

            return GetSql(className + "." + method, parameterList);
        }

        #endregion

        /// <summary>
        /// Returns the SQL string associated with a specific key.
        /// </summary>
        /// <param name="className">Class name</param>
        /// <param name="method">Function name</param>
        /// <returns>SQL string</returns>
        public static string GetSql(string name)
        {
            return GetSql(name, new List<string>());
        }

        /// <summary>
        /// Returns the SQL string associated with a specific key. Passes in parameters.
        /// </summary>
        /// <param name="name">Class name</param>
        /// <param name="parameterList">List of string parameters</param>
        /// <returns>SQL string</returns>
        public static string GetSql(string name, List<string> parameterList)
        {
            if (name.Length == 0)
            {
                throw new Exception("Empty Name found in SqlStringFactory.");
            }

            return InsertParameters(sqlStringFactoryConfiguration.SqlStrings[name].Sql, parameterList, name);
        }

        /// <summary>
        /// Replaces each parameter in the SQL string with a parameter item. Ie., {0} {1} {2}.
        /// If not enough parameters are passed in, an exception is thrown.
        /// If too many parameters are passed in, an exception is thrown.
        /// </summary>
        /// <param name="sql">SQL string, optionally including parameters {0} {1} {2}, etc</param>
        /// <param name="parameterList">List of string parameters</param>
        /// <param name="name">Name (for error logging)</param>
        /// <returns>SQL string</returns>
        private static string InsertParameters(string sql, List<string> parameterList, string name)
        {
            int parameterIndex = 0;

            foreach (string parameter in parameterList)
            {
                string parameterKey = "{" + parameterIndex + "}";

                if (sql.IndexOf(parameterKey) == -1)
                {
                    throw new Exception("Too many parameters passed into GetSql. Name: " + name + ", SQL: " + sql + ", Parameter Count: " + parameterList.Count);
                }

                // Replace the occurance of {0} with parameterList[0].
                sql = sql.Replace(parameterKey, parameter);

                parameterIndex++;
            }

            if (sql.IndexOf("{") != -1)
            {
                throw new Exception("Too few parameters passed into GetSql. Name: " + name + ", SQL: " + sql + ", Parameter Count: " + parameterList.Count);
            }

            return sql;
        }
    }
}
